// WordSntxAddn.cpp : Implementation of CWordSntxAddn
//
// Copyright 2006 (c) RegExLab.com
//
// Author: ʷΰ (sswater shi)
//
// 2006/05/19 01:52:42
//

#include "stdafx.h"
#include "SntxAddn.h"
#include "WordSntxAddn.h"

#include "SyntaxModule.h"
#include "ProgressThread.h"

#include "BtnFaceUtil.h"
#include "ClipboardBackup.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

/////////////////////////////////////////////////////////////////////////////
// CWordSntxAddn

_ATL_FUNC_INFO CWordSntxAddn::OnClickButtonInfo = { CC_STDCALL, VT_EMPTY, 2, { VT_DISPATCH, VT_BYREF | VT_BOOL } };

CWordSntxAddn::CWordSntxAddn()
{
	m_spApp = NULL;

	for( int i=0; i<MAX_BUTTONS; i++ )
	{
		m_spButtons[i] = NULL;
		m_pSyntax  [i] = NULL;
	}
}

STDMETHODIMP CWordSntxAddn::InterfaceSupportsErrorInfo(REFIID riid)
{
	static const IID* arr[] = 
	{
		&IID_IWordSntxAddn
	};
	for (int i=0; i < sizeof(arr) / sizeof(arr[0]); i++)
	{
		if (InlineIsEqualGUID(*arr[i],riid))
			return S_OK;
	}
	return S_FALSE;
}

STDMETHODIMP CWordSntxAddn::OnConnection(IDispatch * Application, ext_ConnectMode ConnectMode, IDispatch * AddInInst, SAFEARRAY * * custom)
{
	// Application
	m_spApp = CComQIPtr <_Application> (Application);

	// Command Bar
	CComPtr <CommandBar> spNewCmdBar = InitCommandBar( m_spApp );

	// Load Syntax
	LoadSyntaxes();

	// Buttons
	BOOL bUpdated = InitCommandButtons( spNewCmdBar );

	// Show or Hide
	if( m_pSyntax[0] != NULL )
	{
		if( bUpdated )
		{
			spNewCmdBar->PutVisible(VARIANT_TRUE);
		}
	}
	else
	{
		spNewCmdBar->PutVisible(VARIANT_FALSE);
	}

	return S_OK;
}

STDMETHODIMP CWordSntxAddn::OnDisconnection(ext_DisconnectMode RemoveMode, SAFEARRAY * * custom)
{
	int i;

	for(i=0; i<MAX_BUTTONS; i++)
	{
		if( m_spButtons[i] != NULL )
		{
			DispEventUnadvise(i+1, (IDispatch*)m_spButtons[i]);
			m_spButtons[i] = NULL;
		}

		if( m_pSyntax[i] != NULL )
		{
			delete m_pSyntax[i];
			m_pSyntax[i] = NULL;
		}
	}

	if( RemoveMode == ext_dm_UserClosed )
	{
		RemoveCommandBar( m_spApp );
	}

	return S_OK;
}

VOID __stdcall CWordSntxAddn::OnButtonClick(IDispatch * Ctrl, VARIANT_BOOL * CancelDefault)
{
	AFX_MANAGE_STATE(AfxGetStaticModuleState());

	USES_CONVERSION;

	// check document count
	if( m_spApp->GetWindows()->GetCount() == 0 )
		return;

	// check selection type
	CComPtr <Selection> selection = m_spApp->GetSelection();
	WdSelectionType     seltype   = selection->GetType();

	if( seltype < wdSelectionNormal )
	{
		// no selection
		return;
	}
	else if( seltype > wdSelectionNormal )
	{
		// invalid selection
		TCHAR info[20];
		_tcscpy(info, RSTR(IDS_INFO));
		MessageBox(NULL, RSTR(IDS_SEL_INVALID), info, MB_OK);
		return;
	}

	CComQIPtr <_CommandBarButton> spButton( Ctrl );
	_bstr_t tag = spButton->GetTag();

	CSyntaxModule * pSyntax = NULL;

	for( int i=0; i<MAX_BUTTONS; i++ )
	{
		if( m_pSyntax[i] == NULL )
			break;

		if( m_pSyntax[i]->m_szSyntaxName == tag )
		{
			pSyntax = m_pSyntax[i];
			break;
		}
	}

	if( pSyntax == NULL )
	{
		// invalid
		return;
	}

	// infomation
	int start = selection->Start;
	int end   = selection->End;
	int len   = end - start;

	// reset color
	selection->Font->PutColor( wdColorAutomatic );

	// reset bgcolor
	selection->Font->Shading->PutTexture( wdTextureNone );
	selection->Font->Shading->PutForegroundPatternColor( wdColorAutomatic );
	selection->Font->Shading->PutBackgroundPatternColor( wdColorAutomatic );

	// get text
	_bstr_t bstr_text = selection->GetText();
	LPWSTR  text      = bstr_text;

	// unselect
	selection->SetRange( end, end );

	// parse
	pSyntax->Parse( text, len );

	// show progress
	CProgressDialog dlg;

	dlg.m_nLength = len;

	CProgressThread * pThread = (CProgressThread*)AfxBeginThread(RUNTIME_CLASS(CProgressThread),
        THREAD_PRIORITY_NORMAL,
        0,
        CREATE_SUSPENDED,NULL);

	pThread->m_pDialog = &dlg;

	pThread->ResumeThread();

	// highlight
	CComPtr <_Document> doc = m_spApp->GetActiveDocument();
	int relative_pos, relative_len;

	while( dlg.m_bRunning && pSyntax->GetNextPos(relative_pos, relative_len) )
	{
		// validation
		if( relative_pos >= len )
			continue;

		if( (relative_pos + relative_len) > len )
			relative_len = len - relative_pos;

		// get range
		VARIANT range_start = { VT_I4 }, range_end = { VT_I4 };

		range_start.intVal = start + relative_pos;
		range_end  .intVal = start + relative_pos + relative_len;

		CComPtr <Range> range = doc->Range(&range_start, &range_end);

		int color;

		// fore color
		if( pSyntax->GetColor( color ) )
		{
			color &= 0xFFFFFF;

			range->Font->PutColor( (WdColor)color );
		}

		// back color
		if( pSyntax->GetBgColor( color ) )
		{
			color &= 0xFFFFFF;

			long hol = 0; // head of line
			long eol = 0; // end  of line

			long relative_start = start + relative_pos;

			LPCWSTR sz = text + relative_pos;

			for( eol=0; eol<=relative_len; eol++ )
			{
				if( eol==relative_len || (sz[eol] >= 0x0A && sz[eol] <= 0x0D) )
				{
					if( hol < eol )
					{
						VARIANT range2_start = { VT_I4 }, range2_end = { VT_I4 };

						range2_start.intVal = relative_start+hol;
						range2_end  .intVal = relative_start+eol;

						doc->Range( &range2_start, &range2_end )->Shading->PutForegroundPatternColor( (WdColor)color );
					}

					hol = eol + 1;
				}
			}
		}

		// Progress
		dlg.SetProgress(relative_pos + relative_len);
	}

	// close progress
	dlg.m_synCreateProgress.Lock();

	if( dlg.m_bRunning && dlg.m_bInitalized )
		dlg.SendMessage(WM_CLOSE);

	dlg.m_bRunning = FALSE;

	dlg.m_synCreateProgress.Unlock();

	WaitForSingleObject(pThread->m_hThread, INFINITE);
}

HRESULT CWordSntxAddn::DispEventAdvise  ( int index, IDispatch * pDisp )
{
	switch( index )
	{
		case 1: return ButtonEvents1::DispEventAdvise( pDisp );
		case 2: return ButtonEvents2::DispEventAdvise( pDisp );
		case 3: return ButtonEvents3::DispEventAdvise( pDisp );
		case 4: return ButtonEvents4::DispEventAdvise( pDisp );
		case 5: return ButtonEvents5::DispEventAdvise( pDisp );
		case 6: return ButtonEvents6::DispEventAdvise( pDisp );
		case 7: return ButtonEvents7::DispEventAdvise( pDisp );
		case 8: return ButtonEvents8::DispEventAdvise( pDisp );
		case 9: return ButtonEvents9::DispEventAdvise( pDisp );
	}

	return 0;
}

HRESULT CWordSntxAddn::DispEventUnadvise( int index, IDispatch * pDisp )
{
	switch( index )
	{
		case 1: return ButtonEvents1::DispEventUnadvise( pDisp );
		case 2: return ButtonEvents2::DispEventUnadvise( pDisp );
		case 3: return ButtonEvents3::DispEventUnadvise( pDisp );
		case 4: return ButtonEvents4::DispEventUnadvise( pDisp );
		case 5: return ButtonEvents5::DispEventUnadvise( pDisp );
		case 6: return ButtonEvents6::DispEventUnadvise( pDisp );
		case 7: return ButtonEvents7::DispEventUnadvise( pDisp );
		case 8: return ButtonEvents8::DispEventUnadvise( pDisp );
		case 9: return ButtonEvents9::DispEventUnadvise( pDisp );
	}

	return 0;
}

BOOL CWordSntxAddn::LoadSyntaxes()
{
	// build filespec
	TCHAR path[_MAX_PATH];
	GetModuleFileName(_Module.GetModuleInstance(), path, _MAX_PATH);

	int pos = _tcslen(path) - 1;
	while(pos >= 0 && path[pos] != '\\' && path[pos] != '/') pos --;

	_tcscpy(path + pos + 1, _T("*.dll"));

	// find
	_tfinddata_t fileinfo;
	long handle, result;
	int  count = 0;

	for( result = handle = _tfindfirst(path, &fileinfo); result != -1; result = _tfindnext(handle, &fileinfo) )
	{
		_tcscpy(path + pos + 1, fileinfo.name);
		m_pSyntax[count] = CSyntaxModule::LoadSyntaxModule(path);

		if( m_pSyntax[count] != NULL )
		{
			m_pSyntax[count]->m_szSyntaxName = CString("syntax_mod_") + fileinfo.name;
			count ++;
		}

		if( count >= MAX_BUTTONS )
			break;
	}

	if( handle != -1 )
		_findclose(handle);

	return m_pSyntax[0] != NULL;
}

VOID CWordSntxAddn::RemoveCommandBar(CComPtr <_Application> & wordApp)
{
	CComPtr <_CommandBars> spCmdBars = wordApp->GetCommandBars();
	CComPtr <CommandBar> spSyntaxCmdBar = NULL;
	
	HRESULT hr = spCmdBars->get_Item(_variant_t(RSTR(IDS_SYNTAX_BAR)), &spSyntaxCmdBar);

	if( !FAILED(hr) && spSyntaxCmdBar != NULL )
	{
		spSyntaxCmdBar->Delete();

		// Save
		CComQIPtr <Template> custom( wordApp->GetCustomizationContext() );
		custom->Save();
	}
}

CComPtr <CommandBar> CWordSntxAddn::InitCommandBar(CComPtr <_Application> & wordApp)
{
	CComPtr <_CommandBars> spCmdBars = wordApp->GetCommandBars();
	CComPtr <CommandBar> spNewCmdBar = NULL;

	// attach old created
	HRESULT hr = spCmdBars->get_Item(_variant_t(RSTR(IDS_SYNTAX_BAR)), &spNewCmdBar);

	if( FAILED(hr) || spNewCmdBar == NULL )
	{
		// New Create
		CComVariant vName (RSTR(IDS_SYNTAX_BAR));
		CComVariant vPos  (msoBarFloating);
		CComVariant vTemp (VARIANT_FALSE); // menu is NOT temporary 
		CComVariant vEmpty(DISP_E_PARAMNOTFOUND, VT_ERROR); 

		spNewCmdBar = spCmdBars->Add(vName, vPos, vEmpty, vTemp);
	}

	return spNewCmdBar;
}

BOOL CWordSntxAddn::InitCommandButtons(CComPtr <CommandBar> & spNewCmdBar)
{
	int nAttachIndex = 0, i;

	CComPtr <CommandBarControls> spBarControls = spNewCmdBar->GetControls();

	//MsoControlType::msoControlButton = 1
	CComVariant vToolBarType(1);
	CComVariant vEmpty(DISP_E_PARAMNOTFOUND, VT_ERROR); 

	BOOL bUpdated = FALSE;

	// attach buttons
	int btncnt = spBarControls->GetCount();
	for(i=btncnt; i>=1; i--)
	{
		CComPtr <CommandBarControl> spNewBar = spBarControls->GetItem(_variant_t((long)i));

		CComQIPtr <_CommandBarButton> spButton( spNewBar );
		_bstr_t tag = spButton->GetTag();

		CSyntaxModule * pSyntax = NULL;

		for(int j=0; j<MAX_BUTTONS; j++)
		{
			if( m_pSyntax[j] == NULL )
				break;

			if( m_pSyntax[j]->m_szSyntaxName == tag )
			{
				pSyntax = m_pSyntax[j];
				break;
			}
		}

		if( pSyntax != NULL )
		{
			// keep reference to button disp
			m_spButtons[nAttachIndex ++] = spButton;
			DispEventAdvise(nAttachIndex, (IDispatch*)spButton);

			pSyntax->m_bAttached = TRUE;
		}
		else
		{
			spButton->Delete();
			bUpdated = TRUE;
		}
	}

	// Default Bitmap
	HBITMAP hDefaultFace =::LoadBitmap(_Module.GetResourceInstance(), MAKEINTRESOURCE(IDB_HIGHLIGHT));

	CClipboardBackup cbbk;

	for(i=0; i<MAX_BUTTONS; i++)
	{
		if( m_pSyntax[i] == NULL )
			break;

		if( ! m_pSyntax[i]->m_bAttached )
		{
			// backup if not backup
			if( ! cbbk.m_bFilled )
			{
				cbbk.Backup();
			}

			// create
			CComPtr <CommandBarControl> spNewBar = spBarControls->Add(vToolBarType,vEmpty,vEmpty,vEmpty,vEmpty); 
			ATLASSERT(spNewBar);

			CComQIPtr <_CommandBarButton> spButton (spNewBar);
			ATLASSERT(spButton);

			// set
			spButton->PutStyle(Office::msoButtonIconAndCaption);

			HBITMAP bmBtnFace = m_pSyntax[i]->GetBtnFace();
			if( bmBtnFace != NULL )
			{
				CBtnFaceUtil::PasteTransBitmap(bmBtnFace, CComPtr <_CommandBarButton> (spButton));
				::DeleteObject(bmBtnFace);
			}
			else
			{
				CBtnFaceUtil::PasteTransBitmap(hDefaultFace, CComPtr <_CommandBarButton> (spButton));
			}

			spButton->PutCaption(m_pSyntax[i]->GetCaption()); 
			spButton->PutTooltipText(m_pSyntax[i]->GetTooltip());
			spButton->PutEnabled(VARIANT_TRUE);
			spButton->PutVisible(VARIANT_TRUE); 

			// enable tag
			spButton->PutTag((LPCTSTR)m_pSyntax[i]->m_szSyntaxName);

			// keep reference to button disp
			m_spButtons[nAttachIndex ++] = spButton;
			DispEventAdvise(nAttachIndex, (IDispatch*)spButton);

			bUpdated = TRUE;
		}
	}

	cbbk.Restore();

	::DeleteObject(hDefaultFace);

	return bUpdated;
}
